#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pud.h"

#define DICTSIZE 4096

char dict[DICTSIZE*2];
long lumpsize;

struct
{ char name[32];
  char filename[13];
  long offset;
  long length;
} pudindex[1024];

int indexp=0;

void error(char *fmt, ...)
{ va_list args;

  va_start(args, fmt);
  vprintf(fmt, args);
  va_end(args);
  exit(1);
}

void initpudfile(char *fn)
{ FILE *f;
  char buf[4];

  printf("Initializing \"%s\"...", fn);
  fflush(stdout);
  f=fopen(fn, "rb");
  if(!f)
    error("couldn't open\n");
  fread(buf, 4, 1, f);
  if(memcmp(buf, "IPUD", 4))
    error("not a pud file\n");
  fread(buf, 4, 1, f);
  fseek(f, *(long *)(buf), SEEK_SET);
  while(!feof(f))
  { fread(pudindex[indexp].name, 32, 1, f);
    fread(&(pudindex[indexp].offset), 4, 1, f);
    fread(&(pudindex[indexp].length), 4, 1, f);
    strcpy(pudindex[indexp].filename, fn);
    indexp++;
  }
  fclose(f);
  indexp--;
  printf("done\n");

#ifdef DEBUG
  for(i=0; i<indexp; i++)
    printf("%s:%s/%d/%d\n", pudindex[i].name, pudindex[i].filename,
                            pudindex[i].offset, pudindex[i].length);
#endif
}

char *lzdecompress(char *src, int size)
{ char flag=255;
  char *dest;
  int ds=0, de=DICTSIZE-1;
  int i, dsize=0, n, p;

  memset(dict, 0, DICTSIZE*2);

  lumpsize=*(int*)src;
  dest=(char *)malloc(*(int*)src);
  if(!dest) error("Out of memory decompressing.");
  src+=4;
  size-=4;

  for(i=0; i<size; i++)
  { if(src[i]!=flag) 
    { dest[dsize++]=src[i];
      dict[de++]=src[i];
      ds++;
    }
    else if(!src[i+1])
    { dest[dsize++]=flag;
      dict[de++]=flag;
      ds++;
      i++;
    }
    else
    { n=src[i+1];
      p=src[i+2]+src[i+3]*256;
      i+=3;
      memcpy(dest+dsize, dict+ds+p, n);
      if(n+de>DICTSIZE*2-1)
      { memcpy(dict, dict+DICTSIZE, DICTSIZE);
        de-=DICTSIZE, ds-=DICTSIZE;
      }
      memcpy(dict+de, dest+dsize, n);
      de+=n, ds+=n, dsize+=n;
    }
    if(de>DICTSIZE*2-1)
    { de-=DICTSIZE, ds-=DICTSIZE;
      memcpy(dict, dict+DICTSIZE, DICTSIZE);
    }
  }

  return dest;
}

char *loadlump(char *s)
{ int i, t;
  FILE *f;
  char *lump;
  char *lump2;

  for(i=0, t=-1; i<indexp; i++)
    if(!strcmp(pudindex[i].name, s)) t=i;
  if(t==-1)
    error("Couldn't find lump: \"%s\"\n", s);
  lump=(char *)malloc(pudindex[t].length);
  if(!lump)
    error("Couldn't allocate %li bytes for lump \"%s\"\n", pudindex[t].length, s);
  f=fopen(pudindex[t].filename, "rb");
  fseek(f, pudindex[t].offset, SEEK_SET);
  fread(lump, pudindex[t].length, 1, f);
  fclose(f);

  lump2=lzdecompress(lump, pudindex[t].length);
  free(lump);

  return lump2;
}
